/* Socket server (and client!) utilities, */
/* intended to simplify the porting of JNOS servers and clients to Unix */
/* N2RJT - Dave Brown */
/* Version 1.00 9/29/94 - first released version */

#include <stdio.h>
#include <sys/types.h>
#ifdef __linux__
#define UNIX
#include <unistd.h>
#else
#include <sys/select.h>
#endif
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <memory.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include "sockserv.h"

/* For Linux, I don't know why */
#ifndef FD_SET
#define FD_SET __FD_SET
#endif
#ifndef FD_CLR
#define FD_CLR __FD_CLR
#endif
#ifndef FD_ISSET
#define FD_ISSET __FD_ISSET
#endif
#ifndef FD_ZERO
#define FD_ZERO __FD_ZERO
#endif
#ifndef FD_SETSIZE
#define FD_SETSIZE __FD_SETSIZE
#endif

/* This structure holds the buffers for each open socket.  It was an */
/* afterthought, so it is not used for anything else.  The array is only */
/* allocated when the socket is opened, and freed when the socket is closed. */
struct sockbuffer {
    char *buf;
    int buflen;
    int fragment;
} sockbuf[FD_SETSIZE];

/* This array holds the SERVED sockets, that can be connected to. */
/* It does not hold CLIENT sockets. */
int lsock[MAX_SERVED_SOCKETS];
int nlsock = 0;
struct timeval *socktimeval = NULL;
struct timeval *selecttimeval = NULL;

fd_set readfds, openfds;
int nfds = 0;
int ifds = -1;

#define SOBUF 512
#define NULLCHAR (char *) NULL

int close_s(int s)
{
    FD_CLR(s,&openfds);
    free (sockbuf[s].buf);
    sockbuf[s].buflen = 0;
    sockbuf[s].fragment = 0;
    close(s);
}

void fds_copy(tofd, fmfd)
fd_set *tofd;
fd_set *fmfd;
{
    memcpy(tofd,fmfd,sizeof(fd_set));
}

int usputs(int s,char *buf)
{
    int len;
    len = strlen(buf);
    if (write(s, buf, len) < 0)
        return -1;
    else
        return len;
}

int usvprintf(int s,char *fmt,va_list args)
{
    int len,withargs;
    char *buf;
    
    if (strchr(fmt,'%')==NULLCHAR) {
        /* Common case optimization: no args */
        withargs = 0;
        buf = fmt;
        len = strlen(fmt);
    } else {
        /* Use a default value that is huge */
        withargs = 1;
        buf = (char *) malloc(SOBUF);
        if((len=vsprintf(buf,fmt,args)) >= SOBUF) {
            /* It's too late to be sorry.  He's dead, Jim */
            fprintf(stderr,"usprintf() exceeded %d bytes (%d bytes)\n",
              SOBUF, len);
            exit(1);
        }
    }
    len = usputs(s,buf);
    if(withargs)
        free(buf);
    return len;
}

int usprintf(int s,char *fmt,...)
{
    va_list args;
    int len;
    
    va_start(args,fmt);
    len = usvprintf(s,fmt,args);
    va_end(args);
    return len;
}

int tprintf(char *fmt,...)
{
    va_list args;
    int len;
    
    va_start(args,fmt);
    len = usvprintf(ifds,fmt,args);
    va_end(args);
    return len;
}

int tputstr(char *buf)
{
    int len;
    len = strlen(buf);
    if (write(ifds, buf, len) < 0)
        return -1;
    else
        return len;
}

int tputc(char c)
{
    int len, ic;
    len = 1;
    ic = c;
    if (write(ifds,&ic,len) < 0)
        return -1;
    else
        return len;
}

static void (*login[MAX_SERVED_SOCKETS]) (int i);

int startup(int portnum, void (*newin)(int))
{   int i;
    struct sockaddr_in sin;

    /* First-time initialization */
    if (!nlsock) {
        for (i = 0; i<FD_SETSIZE; i++){
            sockbuf[i].buf = NULL;
            sockbuf[i].buflen = 0;
            sockbuf[i].fragment = 0;
        }
        FD_ZERO(&openfds);
    }

    while ((lsock[nlsock] = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
        if (errno != EINTR) {
            perror("startup: socket");
            exit(1);
        }
    }
    
    bzero(&sin, sizeof(sin));
    sin.sin_family = AF_INET;
    sin.sin_addr.s_addr = INADDR_ANY;
    sin.sin_port = htons(portnum);
    setsockopt(lsock[nlsock], SOL_SOCKET, SO_REUSEADDR, (char *)0, 0);
    while (bind(lsock[nlsock], &sin, sizeof(sin)) == -1) {
        if (errno != EINTR) {
            perror("startup: bind");
            exit(1);
        }
    }
    while (listen(lsock[nlsock], 5) ==  -1) {
        if (errno != EINTR) {
            perror("startup: listen");
            exit(1);
        }
    }

    login[nlsock] = newin;
    FD_SET(lsock[nlsock],&openfds);
    sockbuf[lsock[nlsock]].buf = (struct sockbuffer *) malloc (sizeof(char) * SOBUF);
    sockbuf[lsock[nlsock]].buflen = 0;
    sockbuf[lsock[nlsock]].fragment = 0;
    if (nfds <= lsock[nlsock])
        nfds = lsock[nlsock]+1;
    if (ifds == -1)
        ifds = nfds-1;
    nlsock++;
}

int recvline(int *fd, char *buf, int buflen)
{
    int len,ns,i;
    struct sockaddr_in client;
    char *nl;

    if (selecttimeval != NULL && socktimeval != NULL) {
        selecttimeval->tv_sec = socktimeval->tv_sec;
        selecttimeval->tv_usec = socktimeval->tv_usec;
    }
    if (ifds == -1)
        ifds = nfds-1;
    while (1) {
        if (++ifds == nfds) {
            fds_copy(&readfds, &openfds);
            while ((ifds = select(nfds, &readfds, (fd_set *)NULL, (fd_set *)NULL, 
              selecttimeval)) < 0) {
                if (errno != EINTR) {
                    perror("recvline: select");
                    exit(1);
                }
                fds_copy(&readfds, &openfds);
            }
            if (!ifds)
                return -2;
            ifds = 0;
        } else if (FD_ISSET(ifds,&readfds)) {
            for (i=0; i<nlsock; i++) 
                if (lsock[i] == ifds)
                    break;
            if (i<nlsock) {
                if (FD_ISSET(lsock[i],&readfds)) {
                    len = sizeof(client);
                    while ((ns = accept(lsock[i], &client, &len)) == -1) {
                        if (errno != EINTR) {
                            perror("recvline: accept");
                            exit(1);
                        }
                    }
                    if (nfds<=ns)
                        nfds = ns+1;
                    FD_SET(ns,&openfds);
                    sockbuf[ns].buf = (char *) malloc (sizeof(char) * SOBUF);
                    sockbuf[ns].buflen = 0;
                    sockbuf[ns].fragment = 0;
                    (*login[i])(ns);
                }
                FD_CLR(lsock[i],&readfds);
            } else {
                if (!sockbuf[ifds].buflen) {
                    while ((sockbuf[ifds].buflen = 
                    read(ifds, sockbuf[ifds].buf+sockbuf[ifds].fragment, SOBUF)) == -1) {
                        if (errno != EINTR) {
                            break;
                        }
                    }
                    if (sockbuf[ifds].buflen <= 0) {
                        FD_CLR(ifds,&openfds);
                        free(sockbuf[ifds].buf);
                        sockbuf[ifds].buflen = 0;
                        *fd = ifds;
                        buf[0] = '\0';
                        return -1;
                    } else {
                        sockbuf[ifds].buflen += sockbuf[ifds].fragment;
                        sockbuf[ifds].fragment = 0;
                        sockbuf[ifds].buf[sockbuf[ifds].buflen] = '\0';
                    }
                }
                if ((nl = strchr(sockbuf[ifds].buf,'\n'))==NULL) {
                    sockbuf[ifds].fragment = sockbuf[ifds].buflen;
                    sockbuf[ifds].buflen = 0;
                    continue;
                } else {
                    *nl = '\0';
                    strcpy(buf,sockbuf[ifds].buf);
                    strcpy(sockbuf[ifds].buf,++nl);
                    len = strlen(buf);
                    sockbuf[ifds].buflen = strlen(sockbuf[ifds].buf);                
                    *fd = ifds;
                    if (sockbuf[ifds].buflen)
                        ifds--;
                }
                return len;
            }
        }
    }    
}

long resolve(char *hostname)
{
    struct hostent *hp;
    unsigned long int haddr;

    if ((hp = gethostbyname(hostname)) == 0) {
        herror("resolve: gethostbyname");
        return -1;
    }
    memcpy(&haddr,hp->h_addr,hp->h_length);
    return(haddr);
}

int startcli(hostname, portnum)
char *hostname;
int portnum;
{
    struct hostent *hp;
    unsigned long int haddr;
    
    if ((hp = gethostbyname(hostname)) == 0) {
        herror("startcli: gethostbyname");
        return -1;
    }
    memcpy(&haddr,hp->h_addr,hp->h_length);
    return startcliaddr(hp->h_addrtype, haddr, portnum);
}

int startcliaddr(family,addr,portnum)
int family;
unsigned long int addr;
unsigned short int portnum;
{   int s,i;
    struct sockaddr_in sin;

    bzero(&sin, sizeof(sin));
    sin.sin_addr.s_addr = addr;
    sin.sin_family = family;
    sin.sin_port = htons(portnum);
    
    while ((s = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
        if (errno != EINTR) {
            perror("startcli: socket");
            return -1;
        }
    }
    
    while (connect(s, (struct sockaddr *)&sin, sizeof(sin)) == -1) {
        if (errno != EINTR) {
            perror("startcli: connect");
            return -1;
        }
    }
    
    /* First-time initialization */
    if (!nlsock) {
        for (i = 0; i<FD_SETSIZE; i++){
            sockbuf[i].buf = NULL;
            sockbuf[i].buflen = 0;
        }
        FD_ZERO(&openfds);
    }
    if (nfds<=s)
        nfds = s+1;
    if (ifds == -1)
        ifds = nfds-1;
    FD_SET(s,&openfds);
    sockbuf[s].buf = (char *) malloc (sizeof(char) * SOBUF);
    sockbuf[s].buflen = 0;
    return s;
}

void socktimeout(int msec)
{
    socktimeval = (struct timeval *) malloc (sizeof(struct timeval));
    selecttimeval = (struct timeval *) malloc (sizeof(struct timeval));
    socktimeval->tv_sec = 0;
    socktimeval->tv_usec = msec*1000L;
}

void nosocktimeout(void)
{
    free(socktimeval);
    socktimeval = NULL;
    free(selecttimeval);
    selecttimeval = NULL;
}
